package org.expeditee.io;

import java.awt.Color;
import java.awt.Font;
import java.awt.Point;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import org.expeditee.items.Item;
import org.expeditee.items.Text;

/**
 * This class provides various methods for converting values to\from Java
 * objects and KMS file values.
 * 
 * @author jdm18
 * 
 */
public class Conversion {

	private static final double FONT_SCALE = 0.85;

	// used for rounding when converting between Sandbox and KMS font sizes.
	private static final double FONT_ROUNDING = 0.07;

	// border around edges
	public static final int X_ADJUST = 5;

	public static final int Y_ADJUST = 20;

	/**
	 * Returns the Color corresponding to the given KMS color code. For example:
	 * <br>
	 * green4 = 0% red, 40% green, 0% blue.
	 * 
	 * @param colorCode
	 *            The KMS color code to convert
	 * @return The Color object corresponding to the given code
	 */
	public static Color getColor(String colorCode, Color current) {
		if (colorCode == null || colorCode.equals("null")
				|| colorCode.equals("none")) {
			return current;
		}

		// check if its a normal rgb code ie. 100 0 40
		Color rgb = getRGBColor(colorCode, current);
		if (rgb != null)
			return rgb;

		// separate percentage from color (if necessary)
		String num = "";
		int last = colorCode.length() - 1;

		char c = colorCode.charAt(last);

		while (Character.isDigit(c)) {
			num = c + num;
			if (last <= 0)
				break;

			c = colorCode.charAt(--last);
		}

		float amount = 10f;
		if (num.length() > 0)
			amount = Float.parseFloat(num);
		
		if(amount > 10) 
			amount = 10;

		float color[] = { 0, 0, 0 };

		if (colorCode.toLowerCase().startsWith("red"))
			color[0] = amount / 10;
		else if (colorCode.toLowerCase().startsWith("green"))
			color[1] = amount / 10;
		else if (colorCode.toLowerCase().startsWith("blue"))
			color[2] = amount / 10;
		else
			return current;

		return new Color(color[0], color[1], color[2]);
	}

	private static Color getRGBColor(String colorCode, Color current) {
		int color[] = { 0, 0, 0 };

		try {
			String[] values = colorCode.trim().split("\\s+");
			
			String r = values.length > 0 ? values[0] : "0";
			String g = values.length > 1 ? values[1] : "0";
			String b = values.length > 2 ? values[2] : "0";

			if (current != null && (r.startsWith("+") || r.startsWith("-"))) {
				int mult = 1;
				if (r.startsWith("-"))
					mult = -1;

				r = r.substring(1);

				color[0] = combine(r, current.getRed(), mult);
				color[1] = combine(g, current.getGreen(), mult);
				color[2] = combine(b, current.getBlue(), mult);
			} else {
				color[0] = Integer.parseInt(r);
				color[1] = Integer.parseInt(g);
				color[2] = Integer.parseInt(b);

				for (int i = 0; i < color.length; i++)
					color[i] = (int) Math.floor((color[i] / 100.0) * 255);
			}

			for (int i = 0; i < color.length; i++) {
				if (color[i] > 255)
					color[i] = 255;

				if (color[i] < 0)
					color[i] = 0;
			}

			return new Color(color[0], color[1], color[2]);
		} catch (Exception e) {
			return null;
		}
	}

	private static int combine(String toParse, int current, int mult) {
		return ((int) Math.floor((Integer.parseInt(toParse) / 100.0) * 255) * mult)
				+ current;
	}

	/**
	 * Converts the given Color object to the corresponding KMS color code
	 * 
	 * @param color
	 *            The Color to be turned into KMS color code.
	 * @return A String containing the KMS color code, NULL if the color is
	 *         black.
	 */
	public static String getKMSColorCode(Color color) {
		if (color == null)
			return null;

		int r = (int) ((color.getRed() / 255.0) * 100 + 0.5);
		int g = (int) ((color.getGreen() / 255.0) * 100 + 0.5);
		int b = (int) ((color.getBlue() / 255.0) * 100 + 0.5);

		return r + " " + g + " " + b;
	}

	/**
	 * Adjusts the given x,y coordinates. This is done to allow room for the
	 * title bar and menu bar
	 * 
	 * @param x
	 *            The x coordinate
	 * @param y
	 *            The y coordinate
	 * @return The new values with a small margin on the left and space on the
	 *         top
	 */
	public static java.awt.Point adjust(int x, int y) {
		return new Point(x + X_ADJUST, y + Y_ADJUST);
	}

	/**
	 * Adjusts the given Point to allow room for the title bar and menu bar.
	 * 
	 * @param p
	 *            The point to adjust.
	 * @return A new point with the adjusted values
	 */
	public static Point adjust(Point p) {
		return adjust(p.x, p.y);
	}

	/**
	 * Restores KMS coordinates after an adjust() has been performed
	 * 
	 * @param point
	 *            The Point to restore
	 * @return a new Point with the restored values.
	 */
	public static java.awt.Point deAdjust(java.awt.Point point) {
		return new Point(point.x - X_ADJUST, point.y - Y_ADJUST);
	}

	/**
	 * Converts the given Font to a corresponding KMS font code.
	 * 
	 * @param font
	 *            The Font to convert to a code.
	 * @return The KMS font code that corresponds to the given Font.
	 */
	public static String getKMSFontCode(Font font) {
		String code = "t";
		String fontName = font.getFamily().toLowerCase();

		for (int i = 0; i < Text.FONT_WHEEL.length; i++) {
			if (Text.FONT_WHEEL[i].equals(fontName)) {
				code = "" + Text.FONT_CHARS[i];
				break;
			}
		}

		switch (font.getStyle()) {
		case Font.BOLD:
			code += "b";
			break;
		case Font.PLAIN:
			code += "r";
			break;
		case Font.ITALIC:
			code += "i";
			break;
		default:
			code += "p";
			break;
		}

		double dsize = font.getSize() / FONT_SCALE;
		if ((Math.ceil(dsize)) - dsize < FONT_ROUNDING)
			dsize = Math.ceil(dsize);
		else
			dsize = Math.floor(dsize);

		code += (int) dsize;

		return code;
	}

	/**
	 * Converts the given KMS font code to a Font object.<br>
	 * For example: <br>
	 * tr16 = times, roman, 16pt
	 * 
	 * @param fontCode
	 *            The font code to convert to a Font
	 * @return the Font that corresponds to the given font code.
	 */
	public static Font getFont(String fontCode) {
		String code = Text.FONT_WHEEL[0];

		char c = fontCode.charAt(0);
		for (int i = 0; i < Text.FONT_CHARS.length; i++) {
			if (c == Text.FONT_CHARS[i]) {
				code = Text.FONT_WHEEL[i] + '-';
				break;
			}
		}

		switch (fontCode.charAt(1)) {
		case 'r':
			code += "Plain";
			break;
		case 'b':
			code += "Bold";
			break;
		case 'i':
			code += "Italic";
			break;
		case 'p':
			code += "BoldItalic";
			break;
		}

		code += "-";

		Font font = null;

		try {
			int size = Integer.parseInt(fontCode.substring(2));
			double dsize = size * FONT_SCALE;
			if (dsize - ((int) dsize) > FONT_ROUNDING)
				dsize = Math.ceil(dsize);

			code += (int) dsize;

			font = Font.decode(code);
		} catch (NumberFormatException nfe) {
			font = Font.decode(fontCode);
		}

		return font;
	}

	/**
	 * Returns the number portion (Frame Number) of the given Frame name.
	 * 
	 * @param framename
	 *            The Frame name to extract the number from
	 * @return The Frame number
	 */
	public static int getFrameNumber(String framename) {
		String num = null;
		// The framename must end in a digit
		assert (Character.isDigit(framename.charAt(framename.length() - 1)));
		// And start with a letter
		assert (Character.isLetter(framename.charAt(0)));
		// start at the end and find the first non digit char
		for (int i = framename.length() - 2; i >= 0; i--) {
			if (!Character.isDigit(framename.charAt(i))) {
				num = framename.substring(i + 1);
				break;
			}
		}

		try {
			return Integer.parseInt(num);
		} catch (NumberFormatException nfe) {
			return -1;
		}
	}

	/**
	 * Returns the frameset poriton of the given Frame name (frame number
	 * removed) converted to lower case.
	 * 
	 * @param frame
	 *            The full name to extract the Frameset name from
	 * @return the name of the Frameset extracted from the given Frame name.
	 */
	public static String getFrameset(String frame) {
		return getFrameset(frame, true);
	}

	public static String getFrameset(String framename, boolean convertToLower) {
		String set = null;
		assert (Character.isDigit(framename.charAt(framename.length() - 1)));
		// And start with a letter
		assert (Character.isLetter(framename.charAt(0)));
		// start at the end and find the first non digit char
		for (int i = framename.length() - 2; i >= 0; i--) {
			if (!Character.isDigit(framename.charAt(i))) {
				set = framename.substring(0, i + 1);
				break;
			}
		}

		if (convertToLower)
			return set.toLowerCase();

		return set;
	}

	/**
	 * Converts the given KMS justification code into a int corresponding to the
	 * constants defined in Item.
	 * 
	 * @param justCode
	 *            The KMS justification code to convert
	 * @return The resulting int corresponding to one of the constants defined
	 *         in Item
	 */
	public static int getJustification(String justCode) {
		assert (justCode != null);
		justCode = justCode.trim().toLowerCase();
		// as read from file
		if (justCode.length() == 1) {
			switch (justCode.charAt(0)) {
			case 'f':
				return Item.JUSTIFICATION_FULL;
			case 'l':
				return Item.JUSTIFICATION_LEFT;
			case 'r':
				return Item.JUSTIFICATION_RIGHT;
			case 'c':
				return Item.JUSTIFICATION_CENTER;
			}
		}

		// from the user
		if (justCode.equals("center"))
			return Item.JUSTIFICATION_CENTER;
		if (justCode.equals("left"))
			return Item.JUSTIFICATION_LEFT;
		if (justCode.equals("right"))
			return Item.JUSTIFICATION_RIGHT;
		if (justCode.equals("full"))
			return Item.JUSTIFICATION_FULL;

		// default justification
		return Item.JUSTIFICATION_NONE;
	}

	public static String getKMSJustificationCode(int justification) {
		switch (justification) {
		case Item.JUSTIFICATION_CENTER:
			return "C";
		case Item.JUSTIFICATION_FULL:
			return "F";
		case Item.JUSTIFICATION_LEFT:
			return "L";
		case Item.JUSTIFICATION_RIGHT:
			return "R";
		default:
			return null;
		}

	}

	public static Object Convert(Class type, String value) {
		return Convert(type, value, null);
	}

	// Will convert from KMS color values to RGB...
	public static Object Convert(Class type, String value, Object orig) {
		// System.out.println("Orig: " + orig);
		assert (value != null);
		value = value.trim();

		if (type == String.class
				&& String.CASE_INSENSITIVE_ORDER.compare(value, "") == 0)
			return null;
		else if (type == String.class)
			return value;

		if (type == Font.class) {
			return getFont(value);
		}

		if (type.equals(Color.class)) {
			if (value.length() == 0)
				return null;

			try {
				// Find the field and value of colorName
				Field field = Class.forName("java.awt.Color").getField(
						value.toLowerCase());
				return (Color) field.get(null);
			} catch (Exception e) {
				return getColor(value, (Color) orig);
			}
		}

		if (type.equals(int.class)) {
			if (orig instanceof Integer
					&& (value.startsWith("+") || value.startsWith("-"))) {
				value = value.replace("+", "");

				return ((Integer) orig) + Integer.parseInt(value);
			}

			if (value.equals("null"))
				return -1;

			return Integer.parseInt(value);
		}

		if (type.equals(float.class)) {
			if (orig instanceof Float
					&& (value.startsWith("+") || value.startsWith("-"))) {
				value = value.replace("+", "");

				return ((Float) orig) + Float.parseFloat(value);
			}

			return Float.parseFloat(value);
		}

		if (type.equals(double.class)) {
			if (orig instanceof Double
					&& (value.startsWith("+") || value.startsWith("-"))) {
				value = value.replace("+", "");

				return ((Double) orig) + Double.parseDouble(value);
			}

			return Double.parseDouble(value);
		}

		if (type.equals(int[].class)) {
			StringTokenizer st = new StringTokenizer(value, " ");
			int[] param = new int[st.countTokens()];
			for (int i = 0; i < param.length; i++) {
				try {
					param[i] = Integer.parseInt(st.nextToken());
				} catch (Exception e) {
					return null;
				}
			}

			return param;
		}

		if (type.equals(Font.class)) {
			return Conversion.getFont(value);
		}

		if (type.equals(boolean.class)) {
			value = value.toLowerCase().trim();

			if (value.equals("t") || value.equals("true")
					|| value.equals("yes") || value.equals("y")
					|| value.equals(""))
				return true;

			if (value.equals("f") || value.equals("false")
					|| value.equals("no") || value.equals("n"))
				return false;

			return null;
		}

		// return name + ": " + value;
		/*
		 * String[] val = new String[1]; val[0] = value;
		 */
		return null;

	}

	public static Object[] Convert(Method method, String value) {
		return Convert(method, value, null);
	}

	public static Object[] Convert(Method method, String value, Object current) {
		value = value.trim();

		String name = method.getName();
		Class[] types = method.getParameterTypes();
		if (name.equals("setPosition")) {
			Point[] p = new Point[1];
			p[0] = new Point();
			p[0].x = Integer.parseInt(value.substring(0, value.indexOf(" ")));
			p[0].y = Integer.parseInt(value.substring(value.indexOf(" ") + 1));
			p[0] = adjust(p[0]);
			return p;
		}

		if (name.equals("setJustification")) {
			Integer[] just = new Integer[1];
			just[0] = getJustification(value);
			return just;
		}

		if (name.endsWith("Arrow")) {
			if (value.indexOf(" ") < 0)
				return null;

			Integer a = Integer
					.parseInt(value.substring(0, value.indexOf(" ")));
			Double b = Double.parseDouble(value
					.substring(value.indexOf(" ") + 1));

			Object[] vals = new Object[2];
			vals[0] = a;
			vals[1] = b;
			return vals;
		}

		if (types.length == 1 && types[0] == List.class) {
			StringTokenizer st = new StringTokenizer(value, "\n");
			List<String> list = new LinkedList<String>();
			while (st.hasMoreTokens())
				list.add(st.nextToken());

			Object[] vals = new Object[1];
			vals[0] = list;
			return vals;
		}

		if (types.length == 1) {
			Object o[] = new Object[1];
			o[0] = Convert(types[0], value, current);
			return o;
		}
		/*
		 * if(types.length == 1 && types[0] == int.class){ Convert(types[0],
		 * value);
		 * 
		 * Integer[] val = new Integer[1]; val[0] = Integer.parseInt(value);
		 * return val; }
		 * 
		 * if(types.length == 1 && types[0] == float.class){ Float[] val = new
		 * Float[1]; val[0] = Float.parseFloat(value); return val; }
		 * 
		 * if(types.length == 1 && types[0] == Font.class){ Font f =
		 * getFont(value); if(f != null){ Font[] res = new Font[1]; res[0] = f;
		 * return res; }
		 * 
		 * return null; }
		 * 
		 * if(types.length == 1 && types[0] == Color.class){ Color c =
		 * getColor(value, null); if(c != null){ Color[] res = new Color[1];
		 * res[0] = c; return res; }
		 * 
		 * return null; }
		 * 
		 * if(types.length == 1 && types[0] == int[].class){ if(value.length() ==
		 * 0) return null;
		 * 
		 * StringTokenizer st = new StringTokenizer(value, " "); int[] param =
		 * new int[st.countTokens()]; for(int i = 0; i < param.length; i++){
		 * try{ param[i] = Integer.parseInt(st.nextToken()); }catch(Exception
		 * e){ return null; } }
		 * 
		 * Object[] o = new Object[1]; o[0] = param; return o; }
		 */

		// return name + ": " + value;
		String[] val = new String[1];
		val[0] = value;
		return val;
	}

	public static Object ConvertToKMS(Method method, Object output) {
		if (output == null)
			return null;

		if (method == null) {
			System.out.println("Output: " + output);
		}

		String name = method.getName();

		if (output instanceof List)
			return output;

		if (name.endsWith("Justification"))
			return getKMSJustificationCode((Integer) output);

		// strings can be returned immediately
		if (output instanceof String)
			return (String) output;

		// integers can also be returned immediately
		if (output instanceof Integer && (Integer) output >= 0)
			return "" + output;

		// floats can also be returned immediately
		if (output instanceof Float && (Float) output >= 0)
			return "" + output;

		// convert fonts
		if (output instanceof Font)
			return getKMSFontCode((Font) output);

		// convert colors
		if (output instanceof Color)
			return getKMSColorCode((Color) output);

		// covert points
		if (output instanceof Point)
			return (((Point) output).x - X_ADJUST) + " "
					+ (((Point) output).y - Y_ADJUST);

		if (output instanceof Boolean)
			if ((Boolean) output)
				return null;
			else
				return "F";

		if (output instanceof int[]) {
			int[] out = (int[]) output;
			String res = "";
			for (int i : out)
				res += i + " ";

			res = res.trim();
			if (res.length() > 0)
				return res;
		}
		// default
		return null;
	}
}
