package org.expeditee.io;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.expeditee.gui.Frame;
import org.expeditee.items.Constraint;
import org.expeditee.items.InteractiveWidget;
import org.expeditee.items.Item;
import org.expeditee.items.Line;
import org.expeditee.items.Picture;
import org.expeditee.items.WidgetCorner;
import org.expeditee.items.WidgetEdge;
import org.expeditee.stats.SessionStats;

/**
 * Writes a Frame out to a KMS format file.
 * 
 * @author jdm18
 * 
 */
public class ExpWriter extends DefaultFrameWriter {

	private ProxyWriter _writer = null;

	private StringBuilder _stringWriter = null;

	private static final char TERMINATOR = 'Z';

	public ExpWriter() {
		super();
	}

	public void initialise(Frame start) throws IOException {
		String name = start.getFramesetName().toLowerCase();

		if (_filename == null)
			_filename = start.path + name + File.separator
					+ start.getFrameNumber() + ExpReader.EXTENTION;

		_stringWriter = new StringBuilder();
		if (_filename.toLowerCase().equals("clipboard")) {
			_writer = new ProxyWriter(true);
			_filename = "Clipboard";
		} else
			_writer = new ProxyWriter(new FileWriter(_filename));
		
		try {
			_ItemTags.put("S", Item.class.getMethod("getTypeAndID", new Class[] {}));
		} catch (Exception e) {

		}
	}

	/**
	 * Writes the given Frame (and all items it contains) to a KMS file. Note:
	 * File path and location must be set before calling this or it will do
	 * nothing.
	 * 
	 * @param frame
	 *            The Frame to write out to the file.
	 * @throws IOException
	 *             Any exceptions occured by the BufferedWriter.
	 */
	public void outputFrame(Frame frame) throws IOException {
		if (_writer == null)
			return;

		writeHeader(frame);

		// write out each Item in the Frame.
		List<Item> items = frame.getItems();

		// iWidgets are handled specially since 8 items are written as one
		List<InteractiveWidget> seenWidgets = new LinkedList<InteractiveWidget>();

		// do not write items with ID -1, also skip any lines
		for (Item i : items) {

			// Ensure only one of the WidgetCorners represent a single widget
			if (i instanceof WidgetCorner) {
				InteractiveWidget iw = ((WidgetCorner) i).getWidgetSource();
				if (seenWidgets.contains(iw))
					continue;
				seenWidgets.add(iw);
			}

			if (i.getID() >= 0 && !(i instanceof Line)) {
				writeItem(i);
			}
		}

		// write any lines or constraints
		writeTerminator();
		writeLineData();
		writeTerminator();
		writeConstraintData();

		return;
	}

	private void writeHeader(Frame toWrite) throws IOException {
		Iterator<String> it = _FrameTags.keySet().iterator();
		Object[] param = {};

		while (it.hasNext()) {
			String tag = it.next();
			try {
				Object o = _FrameTags.get(tag).invoke(toWrite, param);
				o = Conversion.ConvertToKMS(_FrameTags.get(tag), o);
				if (o != null) {
					writeLine(tag, (String) o);
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		writeTerminator();
	}

	private void writeLine(String tag, String line) throws IOException {
		writeLine(tag + " " + line);
	}

	private void writeTerminator() throws IOException {
		writeLine(TERMINATOR + "\n");
	}

	// writes the given line out to the file
	private void writeLine(String line) throws IOException {
		// do not write empty lines
		if (line == null)
			return;

		String toWrite = line + "\n";

		_writer.write(toWrite);
		_stringWriter.append(toWrite);
	}
	
	// writes the given Item out to the file
	// This method is not used to write out LINE items
	public void writeItem(Item item) throws IOException {
		if (_writer == null)
			return;

		// Dont save the item if it is off the screen
		if (item.offScreen()) {
			//System.out.println("item offscreen!");
			return;
		}

		// Pictures are saved as the corresponding text items that created them
		if (item instanceof Picture)
			item = ((Picture) item).getText();
		else if (item instanceof WidgetCorner)
			item = ((WidgetCorner) item).getWidgetSource().getSource();

		if (item.isLineEnd())
			writePoint(item);
		else if (!(item instanceof Line))
			writeClass(item);
		// lines are saved at the end of the file
		// So dont worry about them in here
		else
			System.out.println("Unknown Item: " + item.getID());

		writeLine("");
	}

	private List<Item> _points = new LinkedList<Item>();

	// Writes out a Dot to the file
	protected void writePoint(Item point) throws IOException {
		_points.add(point);
		writeClass(point);
	}

	// writes out all lines to the file
	private void writeLineData() throws IOException {
		List<Line> seen = new LinkedList<Line>();

		// loop through all points stored
		for (int i = 0; i < _points.size(); i++) {
			List<Line> lines = _points.get(i).getLines();

			// if this point is part of one or more lines
			if (lines != null && lines.size() > 0) {
				for (Line line : lines) {

					// Brook: widget edges are not saved
					if (line instanceof WidgetEdge) {
						seen.add(line);
						continue;
					}

					// only output new lines that have not yet been output
					if (!seen.contains(line)) {
						writeLine("L", line.getID() + " " + line.getLineType());
						writeLine("s", line.getLineEnds());
						writeLine("");

						// add this line to the list of lines that have been
						// seen
						seen.add(line);
					}
				}
			}
		}
	}

	// writes out any constraints to the file
	private void writeConstraintData() throws IOException {
		// outputs any constraints the points have

		// loop through all the points
		while (_points.size() > 0) {
			Item i = _points.get(0);

			// if there are any constraints to write
			if (i.getConstraints() != null) {
				List<Constraint> constraints = i.getConstraints();

				// do not write constraints that have already been
				// written
				for (Constraint c : constraints) {
					if (_points.contains(c.getStart())
							&& _points.contains(c.getEnd())) {
						writeLine("C", c.getID() + " " + c.getType());
						writeLine("s", c.getLineEnds());
						writeLine("");
					}

				}
			}
			// remove the point from the list
			_points.remove(0);
		}
	}

	@Override
	protected String finaliseFrame() throws IOException {
		writeTerminator();

		writeLine(SessionStats.getFrameEventList());

		_writer.flush();
		_writer.close();
		_writer = null;

		return "Frame successfully written to " + _filename;
	}

	private void writeClass(Object toWrite) throws IOException {
		Iterator<String> it = _ItemTags.keySet().iterator();
		Object[] param = {};

		while (it.hasNext()) {
			String tag = it.next();
			Method toRun = _ItemTags.get(tag);
			Class declarer = toRun.getDeclaringClass();
			if (declarer == toWrite.getClass()
					|| declarer == toWrite.getClass().getSuperclass()
					|| declarer == toWrite.getClass().getSuperclass()
							.getSuperclass()) {
				try {
					Object o = toRun.invoke(toWrite, param);
					o = Conversion.ConvertToKMS(toRun, o);
					if (o != null) {
						if (o instanceof List) {
							for (Object line : (List) o)
								writeLine(tag, (String) line);
						} else
							writeLine(tag, (String) o);
					}
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
	}

	/**
	 * Gets a string representation of the frame file contents.
	 */
	public String getFileContents() {
		return _stringWriter.toString();
	}
}
